from tabulate import tabulate

from install_party.util import openstack, ovh


def gather_instances(entries_dict, config):
    """Gather all instances which name belongs to the namespace defined in the
    configuration and add their info to a given dict.

    Args:
        entries_dict (dict): The dict to add the instances' info to.
        config (dict): The parsed configuration.
    """
    nova_client = openstack.get_nova_client(config)

    print("Gathering instances...")

    # Retrieve all instances which name starts with the namespace and is followed by "-".
    instances = nova_client.servers.list(search_opts={
        "name": "%s-*" % config["general"]["namespace"]
    })

    # Edit the entries dictionary to add the instances' information.
    for instance in instances:
        entry_id = instance.name.split("-", 1)[1]
        if entry_id in entries_dict:
            entries_dict[entry_id]["instance"] = instance
        else:
            entries_dict[entry_id] = {"instance": instance}


def gather_domains(entries_dict, config):
    """Gather all domain names which subdomain belongs to the namespace defined in the
    configuration and add their info to a given dict.

    Args:
        entries_dict (dict): The dict to add the domain names' info to.
        config (dict): The parsed configuration.
    """
    ovh_client = ovh.get_ovh_client(config)

    print("Gathering domains...")

    # Retrieve all DNS records which sub domain ends with "." followed by the namespace.
    sub_domain_filter = "%25.{namespace}".format(
        namespace=config["general"]["namespace"],
    )

    record_ids = ovh_client.get(
        "/domain/zone/%s/record?subDomain=%s" % (
            config["general"]["dns_zone"], sub_domain_filter
        )
    )

    # Retrieve data on each sub domain.
    for record_id in record_ids:
        record = ovh_client.get(
            "/domain/zone/%s/record/%s" % (
                config["general"]["dns_zone"], record_id
            )
        )

        # Edit the entries dictionary to add the record's information.
        entry_id = record["subDomain"].split(".", 1)[0]
        if entry_id in entries_dict:
            entries_dict[entry_id]["domain"] = record
        else:
            entries_dict[entry_id] = {"domain": record}


def sort_entries(entries_dict):
    """Process a dict populated by gather_instances and gather_domains and sorts its
    entries into three lists: one containing the entries that have both an instance and a
    domain, one containing those that only have a domain, and one containing those that
    only have an instance.

    All lists are populated in such a way that they can be directly fed to the call to
    tabulate in get_and_print_list.

    Args:
        entries_dict (dict): The dict containing the entries to sort.

    Returns:
        list: The list containing the entries that have both an instance and a domain.
        list: The list containing the entries that only have a domain.
        list: The list containing the entries that only have an instance.
    """

    complete_entries = []
    orphaned_instances = []
    orphaned_domains = []
    for entries_dict, entry in entries_dict.items():
        instance = entry.get("instance")
        domain = entry.get("domain")

        if domain:
            # Generate the full domain name for this entry from the domain's info.
            full_domain = "%s.%s" % (domain["subDomain"], domain["zone"])

        if instance is None:
            # We're sure that domain is not None (and therefore full_domain is defined)
            # here because otherwise this ID wouldn't be in the dict.
            orphaned_domains.append([entries_dict, full_domain, domain["target"]])
        elif domain is None:
            # We're sure that instance is not None here because otherwise this ID wouldn't
            # be in the dict.
            orphaned_instances.append([
                entries_dict,
                instance.name,
                instance.status,
                openstack.get_ipv4(instance)
            ])
        else:
            complete_entries.append([
                entries_dict,
                instance.name,
                full_domain,
                instance.status,
                openstack.get_ipv4(instance)
            ])

    return complete_entries, orphaned_domains, orphaned_instances


def get_and_print_list(config):
    """Retrieve a list of all instances and domain names under a configured namespace and
    print a table listing them and associating each instance with its domain name.

    If an instance doesn't have a domain name associated, or vice-versa, the entry is
    listed in one of two extra tables (depending on what is missing). Each of those extra
    tables is only displayed if it contains at least one entry.

    Args:
        config (dict): The parsed configuration.
    """
    # Initialise the empty dict which will be populated later.
    entries_dict = {}

    # Populate the dict with instances.
    gather_instances(entries_dict, config)
    # Populate the dict with domains.
    gather_domains(entries_dict, config)

    # Sort the entries into three lists.
    complete_entries, orphaned_domains, orphaned_instances = sort_entries(entries_dict)

    # Print the lists.
    print(tabulate(
        complete_entries,
        headers=["Name", "Instance name", "Domain", "Status", "IPv4"],
        tablefmt="psql",
    ))

    if orphaned_instances:
        print("\nORPHANED INSTANCES")
        print(tabulate(
            orphaned_instances,
            headers=["Name", "Instance name", "Status", "IPv4"],
            tablefmt="psql"
        ))

    if orphaned_domains:
        print("\nORPHANED DOMAINS")
        print(tabulate(
            orphaned_domains,
            headers=["Name", "Domain", "Status", "IPv4"],
            tablefmt="psql"
        ))



import ovh


def get_ovh_client(config):
    ovh_config = config["ovh"]
    ovh_client = ovh.Client(
        endpoint=ovh_config["endpoint"],
        application_key=ovh_config["application_key"],
        application_secret=ovh_config["application_secret"],
        consumer_key=ovh_config["consumer_key"],
    )

    return ovh_client



import random
import string
import sys
import time
import os

import requests

from install_party.util import errors, openstack, ovh


def random_string(n):
    """Generate a random string made of n lowercase letters."""
    return ''.join(random.choices(string.ascii_lowercase, k=n))


def create_instance(name, expected_domain, config):
    """Create the instance with a boot script using the OpenStack Nova API, then wait for
    the instance to become active (and raise an exception if an error occurred).

    Args:
        name (str): The suffix for the name of the instance to create. The final name will
            be "namespace-name" where "namespace" is the namespace defined in the
            configuration.
        expected_domain (str): The domain name that is expected to be attached to the
            instance later in the creation process.
        config (dict): The parsed configuration.

    Returns:
        str: the IPv4 address of the instance.

    Raises:
        InstanceCreationError: When waiting for the instance's status to become ACTIVE, it
            instead became ERROR.
    """
    nova_client = openstack.get_nova_client(config)

    print("Creating instance...")

    # Generate the actual script to run post-creation from the template and the
    # configuration.
    post_creation_script_path = os.path.join(sys.prefix, "scripts/post_create.sh")
    post_creation_script = open(post_creation_script_path).read().format(
        user=config["instances"]["user"],
        password=config["instances"]["password"],
        riot_version=config["general"]["riot_version"],
        expected_domain=expected_domain,
    )

    # Ask the hypervisor to create a new instance.
    openstack_config = config["openstack"]
    instance = nova_client.servers.create(
        name="%s-%s" % (config["general"]["namespace"], name),
        image=openstack_config["image_id"],
        flavor=openstack_config["flavor_id"],
        userdata=post_creation_script,
    )

    print("Waiting for instance to become active...")

    # Wait for the instance to become active.
    status = ""
    while status != "ACTIVE":
        instance = nova_client.servers.list(search_opts={
            "name": name,
        })[0]

        status = instance.status

        if status == "ERROR":
            raise errors.InstanceCreationError("The instance status changed to ERROR.")

    return openstack.get_ipv4(instance)


def create_record(name, ip_address, config):
    """Create a DNS A record to attach to an instance using the OVH API.

    Args:
        name (str): The prefix for the DNS record's subdomain. The final subdomain will be
            "name.namespace" where "namespace" is the namespace defined in the
            configuration.
        ip_address (str): The IPv4 address to attach the DNS A record to.
        config (dict): The parsed configuration.

    Returns:
         dict: The parsed JSON response to the record creation request.
    """
    ovh_client = ovh.get_ovh_client(config)

    print("Creating DNS record...")

    # Create a new DNS record in the configured zone.
    record = ovh_client.post(
        "/domain/zone/%s/record" % config["general"]["dns_zone"],
        fieldType="A",
        subDomain="%s.%s" % (name, config["general"]["namespace"]),
        target=ip_address,
    )

    print("Refreshing the DNS zone...")

    # Refresh the DNS server's configuration to make it aware of the new record.
    ovh_client.post("/domain/zone/%s/refresh" % config["general"]["dns_zone"])

    return record


def create(name=None, config={}):
    """Create an instance, attach a domain name to it, and wait until the instance's
    boot script has been run.

    Args:
        name (str): The name of the host (will be used to both name the instance and
            define the domain name).
        config (dict): The parsed configuration.
    """
    # Generate a random name (5 lowercase letters) if none was provided.
    if name is None:
        name = random_string(5)

    # Guess what the final domain name for the host is going to be. This is used for
    # inserting the right values in the post-creation script template.
    expected_domain = "%s.%s.%s" % (
        name, config["general"]["namespace"], config["general"]["dns_zone"]
    )

    print("Provisioning host %s (expected domain name %s)" % (name, expected_domain))
    # Create the instance with the OpenStack API.
    ip_address = create_instance(name, expected_domain, config)
    print("Host is active, IPv4 address is", ip_address)
    # Create a DNS A record for the instance's IP address using OVH's API.
    record = create_record(name, ip_address, config)
    # We use the data the API gave us in response to highlight any possible mismatch
    # between the domain name we guessed and the one we actually created.
    print("Created DNS record %s.%s" % (record["subDomain"], record["zone"]))

    print("Waiting for post-creation script to finish...")
    # Every second, check if we can reach the host's HTTPS server, and only exit it if we
    # got a 200 OK response. Because starting up the HTTP(S) server is the last operation
    # performed by the post-creation script, reaching this condition means that the
    # execution finished successfully.
    while True:
        time.sleep(1)

        try:
            response = requests.get("https://%s" % expected_domain)
        except Exception:
            continue

        if response.status_code == 200:
            break

    print("Host created and provisioned!")



import ipaddress

from novaclient import client as novaclient

# Only imported for type hints.
from novaclient.v2.client import Client as V2Client


def get_nova_client(config) -> V2Client:
    openstack_config = config["openstack"]
    nova_client = novaclient.Client(
        version=openstack_config["api_version"],
        auth_url=openstack_config["auth_url"],
        username=openstack_config["username"],
        password=openstack_config["password"],
        project_id=openstack_config["tenant_id"],
        project_name=openstack_config["tenant_name"],
        region_name=openstack_config["region_name"],
    )

    return nova_client


def get_ipv4(instance):
    # Get the instance's IPv4 address from its metadata.
    # We need to loop through all of the interfaces of the Ext-Net
    # network (which is the public-facing network) because we can't
    # always know how the interfaces are ordered.
    interfaces = instance.addresses.get("Ext-Net")
    if interfaces is None:
        return None

    ip_address = None
    for interface in interfaces:
        address = interface["addr"]
        try:
            # This will raise an AddressValueError exception if the
            # value isn't an IPv4 address.
            ipaddress.IPv4Address(address)
            # If no exception have been raised, then the address is an
            # IPv4 address.
            ip_address = address
        except ipaddress.AddressValueError:
            # This address isn't an IPv4 address; continue to the
            # next interface.
            continue

    return ip_address
